/**
 * Custom Matchbox colours for Material MkDocs
 */

/* Set base colors for light mode with custom primary */
[data-md-color-primary="custom"] {
    --md-primary-fg-color: #1B2E46;
    --md-primary-fg-color--light: hsl(213.5, 44.33%, 39%);
    --md-primary-fg-color--dark: hsl(213.5, 44.33%, 10%);
    --md-typeset-a-color: #1B2E46;
}

/* Override for dark mode with custom primary */
[data-md-color-scheme="slate"][data-md-color-primary="custom"] {
    --md-primary-fg-color: #1B2E46;
    --md-primary-fg-color--light: hsl(213.5, 60%, 70%);
    --md-primary-fg-color--dark: hsl(213.5, 44.33%, 10%);
    --md-typeset-a-color: hsl(213.5, 60%, 70%);
}

[data-md-color-scheme="slate"] .md-button {
    color: hsl(213.5, 60%, 70%);
    border-color: hsl(213.5, 60%, 70%);
}

/* Override accent color for custom setting */
[data-md-color-accent="custom"] {
    --md-accent-fg-color: #E4A242;
    --md-accent-fg-color--transparent: hsla(35.5, 75%, 57.6%, 0.1);
}